#!/usr/bin/env python
from __future__ import absolute_import
import sys

from flask import Flask
import werkzeug.serving

from clay import config

log = config.get_logger('clay.server')


def load_middleware(app, name, mwconfig):
    log.info('Loading WSGI middleware %s' % name)
    try:
        modulename, wsgi = name.rsplit('.', 1)
        mw = __import__(modulename)
        mw = sys.modules[modulename]
        mw = getattr(mw, wsgi, None)
        if mw is None or not callable(mw):
            log.error('No callable named %s in %s (%r)' % (wsgi, modulename, mw))
        else:
            app = mw(app, **mwconfig)
    except Exception, e:
        log.exception('Unable to load WSGI middleware %s' % name)
    return app

flask_init = config.get('flask.init', {
    'import_name': 'clayapp',
})

app = Flask(**flask_init)
app.debug = config.get('debug.enabled', False)
app.config.update(config.get('flask.config', {}))
application = app
for name, mwconfig in config.get('middleware', {}).iteritems():
    application = load_middleware(application, name, mwconfig)


def devserver():
    if not config.get('debug.enabled', False):
        sys.stderr.write('This server must be run in development mode, set debug.enabled in your config and try again\n')
        return -1

    for modulename in config.get('views'):
        log.debug('Loading views from %s' % modulename)
        module = __import__(modulename)

    conf = config.get('debug.server')
    log.warning('DEVELOPMENT MODE')
    log.info('Listening on %s:%i' % (conf['host'], conf['port']))

    kwargs = {
        'use_reloader': True,
        'use_debugger': True,
        'use_evalex': True,
        'threaded': False,
        'processes': 1,
    }
    kwargs.update(config.get('debug.werkzeug', {}))
    werkzeug.serving.run_simple(conf['host'], conf['port'], application, **kwargs)


if __name__ == '__main__':
    sys.exit(devserver())



from __future__ import absolute_import

import logging.config
import logging
import random
import signal
import json
import time
import os.path
import os
import sys

from clay import logger


SERIALIZERS = {'json': json}

try:
    import yaml
    SERIALIZERS['yaml'] = yaml
except ImportError:
    pass


class Configuration(object):
    '''
    Manages global configuration from JSON files
    '''
    def __init__(self):
        self.paths = []
        self.config = {}
        self.last_updated = None
        self.init_logging()

    def load(self, signum=None, frame=None):
        '''
        Called when the configuration should be loaded. May be called multiple
        times during the execution of a program to change or update the
        configuration. This method should be overridden by a subclass.
        '''
        return

    def debug(self):
        '''
        Returns True if this server should use debug configuration and logging.
        This method is deprecated and 
        '''
        log = self.get_logger('clay.config')
        log.warning('Configuration.debug() is deprecated and may be removed in a future release of clay-flask. Please use config.get("debug.enabled", False) instead')
        return self.get('debug.enabled', False)

    def get(self, key, default=None):
        '''
        Get the configuration for a specific variable, using dots as
        delimiters for nested objects. Example: config.get('api.host') returns
        the value of self.config['api']['host'] or None if any of those keys
        does not exist. The default return value can be overridden.
        '''
        value = self.config
        for k in key.split('.'):
            try:
                value = value[k]
            except KeyError:
                return default
        #sys.stderr.write('config: %s=%r\n' % (key, value))
        return value

    def init_logging(self):
        '''
        Configure the default root logger to output WARNING to stderr
        '''
        logging.basicConfig(
            format='%(asctime)s %(name)s %(levelname)s %(message)s',
            level=logging.WARNING)

    def reset_logging(self):
        '''
        Reset the root logger configuration to no handlers
        '''
        root = logging.getLogger()
        if root.handlers:
            for handler in list(root.handlers):
                root.removeHandler(handler)

    def configure_logging(self, log_config):
        '''
        Remove all existing logging configuration and use the given
        configuration instead. The format of the log_config dict is specified at
        http://docs.python.org/2/library/logging.config.html#logging-config-dictschema
        '''
        logging.config.dictConfig(log_config)

    def get_logger(self, name):
        '''
        Returns a Logger instance that may be used to emit messages with the
        given log name, respecting debug behavior.
        '''

        log = logging.getLogger(name)
        if self.get('debug.logging', False):
            log.setLevel(logging.DEBUG)
        else:
            log.setLevel(logging.INFO)
        return log

    def feature_flag(self, name):
        '''
        Returns a boolean value for the given feature, which may be
        probabalistic.
        '''
        feature = self.get('features.%s' % name)
        if not feature:
            return False
        if 'percent' in feature:
            percent = float(feature['percent']) / 100.0
            return (random.random() < percent)
        return feature.get('enabled', False)


class FileConfiguration(Configuration):
    def load(self, signum=None, frame=None):
        '''
        Iterate through expected config file paths, loading the ones that
        exist and can be parsed.
        '''
        cwd = os.getcwd()
        if not cwd in sys.path:
            sys.path.insert(0, cwd)

        self.config = {}
        paths = list(self.paths)

        if 'CLAY_CONFIG' in os.environ:
            paths += os.environ['CLAY_CONFIG'].split(':')

        for path in paths:
            path = os.path.expandvars(path)
            path = os.path.abspath(path)
            config = self.load_from_file(path)
            self.config.update(config)

        self.last_updated = time.time()

        self.init_logging()
        log_config = self.get('logging')
        if log_config:
            self.configure_logging(log_config)

    def load_from_file(self, filename):
        '''
        Attempt to load configuration from the given filename. Returns an empty
        dict upon failure.
        '''
        log = self.get_logger('clay.config')

        try:
            filetype = os.path.splitext(filename)[-1].lstrip('.').lower()
            if not filetype in SERIALIZERS:
                log.warning('Unknown config format %s, parsing as JSON' % filetype)
                filetype = 'json'

            # Try getting a safe_load function. If absent, use 'load'.
            load = getattr(SERIALIZERS[filetype], "safe_load",
                           getattr(SERIALIZERS[filetype], "load"))

            config = load(file(filename, 'r'))
            if not config:
                raise ValueError('Empty config')
            log.info('Loaded configuration from %s' % filename)
            return config
        except ValueError, e:
            log.critical('Error loading config from %s: %s' %
                (filename, str(e)))
            sys.exit(1)
            return {}


CONFIG = FileConfiguration()
CONFIG.load()

# Upon receiving a SIGHUP, configuration will be reloaded
signal.signal(signal.SIGHUP, CONFIG.load)

# Expose some functions at the top level for convenience
get = CONFIG.get
get_logger = CONFIG.get_logger
feature_flag = CONFIG.feature_flag
debug = CONFIG.debug



from __future__ import absolute_import

import raven.utils.wsgi
import raven

from clay import config

log = config.get_logger('clay.sentry')
client = None


def get_sentry_client():
    global client
    if client:
        return client
    dsn = config.get('sentry.url', None)
    if not dsn:
        return
    client = raven.Client(dsn=dsn)
    return client


def exception(exc_info, request=None, event_id=None, **extra):
    try:
        _exception(exc_info, request=request, event_id=event_id, **extra)
    except:
        log.exception('Unable to send event to sentry')


def _exception(exc_info, request=None, event_id=None, **extra):
    client = get_sentry_client()
    if not client:
        # return silently if sentry isn't configured
        return
    if request is not None:
        environ = request.environ
        client.capture('Exception', data={
            'sentry.interfaces.Http': {
                'method': request.method,
                'url': request.base_url,
                'data': request.data,
                'query_string': environ.get('QUERY_STRING', ''),
                'headers': dict(raven.utils.wsgi.get_headers(environ)),
                'env': dict(raven.utils.wsgi.get_environ(environ)),
            },
            'logger': extra.get("logger", "sentry"),
        }, extra=extra, exc_info=exc_info, event_id=event_id)
    else:
        client.captureException(exc_info, extra=extra,
            event_id=event_id)



from clay import config
import threading
import random

log = config.get_logger('clay.database')


class DatabaseContext(object):
    def __init__(self, servers, dbapi_name):
        '''
        Servers is a list of config dicts for connecting to postgres
        '''
        self.servers = servers

        self.tlocal = threading.local()
        self.tlocal.dbconn = None

        if not dbapi_name in ('psycopg2', 'MySQLdb', 'sqlite3'):
            raise NotImplementedError('Unsupported database module: %s' % dbapi_name)
        self.dbapi_name = dbapi_name
        self.dbapi = __import__(dbapi_name)

    def __enter__(self):
        server = random.choice(self.servers)
        conn = self.dbapi.connect(**server)
        self.tlocal.dbconn = conn
        return conn

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.tlocal.dbconn.close()
        self.tlocal.dbconn = None

        if exc_type is not None:
            raise
    
    def __str__(self):
        if self.dbconn is not None:
            return 'DatabaseContext %s %r (connected)' % (self.dbapi_name, self.servers)
        else:
            return 'DatabaseContext %s %r (not connected)' % (self.dbapi_name, self.servers)


read = DatabaseContext(config.get('database.read'), config.get('database.module'))
write = DatabaseContext(config.get('database.write'), config.get('database.module'))



from traceback import format_exc
from datetime import datetime
from Queue import Queue
import threading
import logging
import socket
import json
import time
import ssl


LOG_QUEUE_SIZE = 5000
BACKOFF_INITIAL = 0.1
BACKOFF_MULTIPLE = 1.2
INTERNAL_LOG = logging.getLogger('clay_internal')


class TCPHandler(logging.Handler):
    '''
    Python logging handler for sending JSON formatted messages over
    TCP, optionally wrapping the connection with TLSv1
    '''
    def __init__(self, host, port, ssl_ca_file=None):
        '''
        Instantiate a TCPHandler with the intent of connecting to the
        given host (string) and port (int) with or without using SSL/TLSv1
        '''
        logging.Handler.__init__(self)
        self.host = host
        self.port = port
        self.ssl_ca_file = ssl_ca_file
        self.sock = None
        self.queue = Queue(LOG_QUEUE_SIZE)
        self.connect_wait = BACKOFF_INITIAL
        self.raiseExceptions = 0

        self.hostname = socket.gethostname()
        if self.hostname.find('.') != -1:
            self.hostname = self.hostname.split('.', 1)[0]

        self.sender = threading.Thread(target=self.run)
        self.sender.setDaemon(True)
        self.sender.start()

    def connect(self):
        '''
        Create a connection with the server, sleeping for some
        period of time if connection errors have occurred recently.
        '''
        self.sock = socket.socket()
        if self.ssl_ca_file:
            self.sock = ssl.wrap_socket(self.sock,
                ssl_version=ssl.PROTOCOL_TLSv1,
                cert_reqs=ssl.CERT_REQUIRED,
                ca_certs=self.ssl_ca_file)

        INTERNAL_LOG.debug('Connecting (backoff: %.03f)' %
            self.connect_wait)
        time.sleep(self.connect_wait)
        self.sock.connect((self.host, self.port))

    def jsonify(self, record):
        '''
        Translate a LogRecord instance into a json_event
        '''
        timestamp = datetime.utcfromtimestamp(record.created)
        timestamp = timestamp.isoformat()

        fields = {
            'level': record.levelname,
            'filename': record.pathname,
            'lineno': record.lineno,
            'method': record.funcName,
        }
        if record.exc_info:
            fields['exception'] = str(record.exc_info)
            fields['traceback'] = format_exc(record.exc_info)

        log = {
            '@source_host': self.hostname,
            '@timestamp': timestamp,
            '@tags': [record.name],
            '@message': record.getMessage(),
            '@fields': fields,
        }
        return json.dumps(log)

    def emit(self, record):
        '''
        Send a LogRecord object formatted as json_event via a
        queue and worker thread.
        '''
        self.queue.put_nowait(record)

    def run(self):
        '''
        Main loop of the logger thread. All network I/O and exception handling
        originates here. Strings are consumed from self.queue and sent to
        self.sock, creating a new connection if necessary.

        If any exceptions are caught, the message is put() back on the queue
        and the exception is allowed to propagate up through
        logging.Handler.handleError(), potentially causing this thread to abort.
        '''
        INTERNAL_LOG.debug('Log I/O thread started')
        while True:
            record = self.queue.get()
            if record is None:
                break

            jsonrecord = self.jsonify(record)
            jsonrecord = '%s\n' % jsonrecord

            try:
                if self.sock is None:
                    self.connect()
                self.send(jsonrecord)
            except Exception:
                # This exception will be silently ignored and the message
                # requeued unless self.raiseExceptions=1
                self.queue.put(record)
                self.handleError(record)
            self.queue.task_done()
        INTERNAL_LOG.debug('Log I/O thread exited cleanly')

    def send(self, data):
        '''
        Keep calling SSLSocket.write until the entire message has been sent
        '''
        while len(data) > 0:
            if self.ssl_ca_file:
                sent = self.sock.write(data)
            else:
                sent = self.sock.send(data)
            data = data[sent:]
        self.connect_wait = BACKOFF_INITIAL

    def handleError(self, record):
        '''
        If an error occurs trying to send the log message, close the connection
        and delegate the exception handling to the superclass' handleError,
        which raises the exception (potentially killing the log thread) unless
        self.raiseExceptions is False.
        http://hg.python.org/cpython/file/e64d4518b23c/Lib/logging/__init__.py#l797
        '''
        INTERNAL_LOG.exception('Unable to send log')
        self.cleanup()
        self.connect_wait *= BACKOFF_MULTIPLE
        logging.Handler.handleError(self, record)

    def cleanup(self):
        '''
        If the socket to the server is still open, close it. Otherwise, do
        nothing.
        '''
        if self.sock:
            INTERNAL_LOG.info('Closing socket')
            self.sock.close()
            self.sock = None

    def close(self):
        '''
        Send a sentinel None object to the worker thread, telling it to exit
        and disconnect from the server.
        '''
        self.queue.put(None)
        self.cleanup()
        #self.sender.join()


class UDPHandler(logging.Handler):
    '''
    Python logging handler for sending JSON formatted messages over UDP
    '''
    def __init__(self, host, port):
        '''
        Instantiate a UDPHandler with the intent of connecting to the
        given host (string) and port (int)
        '''
        logging.Handler.__init__(self)
        self.host = host
        self.port = port
        self.sock = None
        self.raiseExceptions = 0

        self.hostname = socket.gethostname()
        if self.hostname.find('.') != -1:
            self.hostname = self.hostname.split('.', 1)[0]

    def connect(self):
        '''
        Create a connection with the server, sleeping for some
        period of time if connection errors have occurred recently.
        '''
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        self.sock.connect((self.host, self.port))

    def jsonify(self, record):
        '''
        Translate a LogRecord instance into a json_event
        '''
        timestamp = datetime.utcfromtimestamp(record.created)
        timestamp = timestamp.isoformat()

        fields = {
            'level': record.levelname,
            'filename': record.pathname,
            'lineno': record.lineno,
            'method': record.funcName,
        }
        if record.exc_info:
            fields['exception'] = str(record.exc_info)
            fields['traceback'] = format_exc(record.exc_info)

        log = {
            '@source_host': self.hostname,
            '@timestamp': timestamp,
            '@tags': [record.name],
            '@message': record.getMessage(),
            '@fields': fields,
        }
        return json.dumps(log)

    def emit(self, record):
        '''
        Send a LogRecord object formatted as json_event via a
        queue and worker thread.
        '''
        try:
            if self.sock is None:
                self.connect()
            jsonrecord = self.jsonify(record)
            jsonrecord = '%s\n' % jsonrecord
            self.sock.sendall(jsonrecord)
        except Exception:
            INTERNAL_LOG.exception('Error sending message to log server')
            self.close()

    def close(self):
        if self.sock:
            self.sock.close()
        self.sock = None



from flask import request
from clay import app, config
import json

log = config.get_logger('clay.docs')


def parse_docstring_param(directive, key, value):
    p = {
        'name': key,
        'description': value.split('{', 1)[0],
        'required': False,
        'dataType': 'string',
        'type': 'primitive',
        'allowMultiple': False,
    }
    if '{' in value and '}' in value:
        p.update(json.loads(value[value.find('{'):value.find('}')]))

    if directive == 'json':
        directive = 'body'
        p['type'] = 'complex'

    if directive in ('query', 'body', 'path', 'form'):
        p['paramType'] = directive
    elif directive == 'reqheader':
        p['paramType'] = 'header'
    else:
        log.warning('Ignoring unknown docstring param %s', directive)
        return
    return p


def parse_docstring(docstring):
    '''
    Turns autodoc http dialect docstrings into swagger documentation
    '''
    if not docstring:
        return
    
    params = []
    responses = []
    stripped = ''
    rtype = None
    for line in docstring.split('\n'):
        line = line.lstrip('\t ')
        if not line.startswith(':'):
            stripped += line + '\n<br />'
            continue
        
        directive, value = line.split(':', 2)[1:]
        value = value.strip('\t ')

        directive = directive.split(' ', 1)
        if len(directive) > 1:
            directive, key = directive
        else:
            directive = directive[0]
            key = None

        if directive in ('json', 'body', 'query', 'path', 'form', 'reqheader'):
            param = parse_docstring_param(directive, key, value)
            if param:
                params.append(param)
            continue

        if directive == 'status':
            responses.append({
                'code': int(key),
                'message': value,
            })
            continue
        
        if directive == 'rtype':
            rtype = value
        log.warning('Ignoring unknown docstring param %s', directive)

    return (params, responses, stripped, rtype)


def get_model(modelspec):
    module, name = modelspec.rsplit('.', 1)
    module = __import__(module)
    return {
        'id': modelspec,
        'properties': getattr(module, name),
    }


@app.route('/_docs', methods=['GET'])
def clay_docs():
    '''
    Returns a JSON document describing this service's API

    Endpoints are inferred from routes registered with Flask and the docstrings
    bound to those methods.

    Dialect documentation http://pythonhosted.org/sphinxcontrib-httpdomain/
    Swagger documentation https://github.com/wordnik/swagger-core/wiki

    :status 200: Generated swagger documentation
    :status 500: Something went horribly wrong
    '''
    headers = {'Content-type': 'application/json'}
    response = {
        'apiVersion': '0.2',
        'swaggerVersion': '1.2',
        'basePath': config.get('docs.base_path', None) or request.url_root.rstrip('/'),
        'resourcePath': '/',
        'apis': [],
        'models': {},
    }

    for rule in app.url_map.iter_rules():
        if rule.endpoint == 'static':
            continue

        api = {
            'path': rule.rule,
            'operations': [],
        }
        view_func = app.view_functions[rule.endpoint]
        if view_func.__doc__:
            docstring = view_func.__doc__.strip('\r\n\t ')
            params, responses, stripped_docstring, rtype = parse_docstring(docstring)

            shortdoc = [x for x in docstring.split('\n') if x]
            if not shortdoc:
                shortdoc = 'Undocumented'
            else:
                shortdoc = shortdoc[0]
        else:
            shortdoc = 'Undocumented'
            params = []
            responses = []
            stripped_docstring = shortdoc
            rtype = None


        for http_method in rule.methods:
            if http_method in ('HEAD', 'OPTIONS'):
                continue
            doc = {
                'method': http_method,
                'nickname': view_func.__name__,
                'summary': shortdoc,
                'notes': stripped_docstring,
                'parameters': params,
                'responseMessage': responses,
            }
            if rtype:
                doc['responseClass'] = rtype
                model = get_model(rtype)
                response['models'][rtype] = model

            api['operations'].append(doc)
        response['apis'].append(api)

    return (json.dumps(response, indent=2), 200, headers)



from __future__ import absolute_import

from flask import make_response
from collections import namedtuple
import contextlib
import functools
import httplib
import urllib2
import urlparse
import os.path
import ssl

from clay import config


Response = namedtuple('Response', ('status', 'headers', 'data'))
log = config.get_logger('clay.http')

DEFAULT_CA_CERTS = '/etc/ssl/certs/ca-certificates.crt'


class VerifiedHTTPSOpener(urllib2.HTTPSHandler):
    def https_open(self, req):
        ca_certs = config.get('http.ca_certs_file', DEFAULT_CA_CERTS)
        if config.get('http.verify_server_certificates', True) and os.path.exists(ca_certs):
            frags = urlparse.urlparse(req.get_full_url())
            ssl.get_server_certificate(
                (frags.hostname, frags.port or 443),
                ca_certs=ca_certs
            )
        return self.do_open(httplib.HTTPSConnection, req)

urllib2.install_opener(urllib2.build_opener(VerifiedHTTPSOpener))


class Request(urllib2.Request):
    '''
    This subclass adds "method" to urllib2.Request
    '''
    def __init__(self, url, data=None, headers={}, origin_req_host=None,
                 unverifiable=False, method=None):
        urllib2.Request.__init__(self, url, data, headers, origin_req_host,
                                 unverifiable)
        if headers is None:
            self.headers = {}
        self.method = method

    def get_method(self):
        if self.method is not None:
            return self.method
        if self.has_data():
            return 'POST'
        else:
            return 'GET'


def request(method, uri, headers={}, data=None, timeout=None):
    '''
    Convenience wrapper around urllib2. Returns a Response namedtuple with 'status', 'headers', and 'data' fields

    It is highly recommended to set the 'timeout' parameter to something sensible
    '''
    req = Request(uri, headers=headers, data=data, method=method)
    if not req.get_type() in ('http', 'https'):
        raise urllib2.URLError('Only http and https protocols are supported')

    try:
        with contextlib.closing(urllib2.urlopen(req, timeout=timeout)) as resp:
            resp = Response(
                status=resp.getcode(),
                headers=resp.headers,
                data=resp.read())
            log.debug('%i %s %s' % (resp.status, method, uri))
    except urllib2.HTTPError, e:
        # if there was a connection error, the underlying fd might be None and we can't read it
        if e.fp is not None:
            resp = Response(
                status=e.getcode(),
                headers=e.hdrs,
                data=e.read())
        else:
            resp = Response(
                status=e.getcode(),
                headers=e.hdrs,
                data=None)
        log.warning('%i %s %s' % (resp.status, method, uri))

    return resp


def cache_control(**cache_options):
    '''
    Decorator that adds a Cache-Control header to the response returned from a
    view. Each keyword argument to this decorator is an option to be appended
    to the Cache-Control header. Underscores '_' are replaced with dashes '-'
    and boolean values are assumed to be directives.

    Examples:
    @cache_control(max_age=0, no_cache=True)
    @cache_control(max_age=3600, public=True)
    '''
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            response = make_response(f(*args, **kwargs))

            cache_control = []
            for key, value in cache_options.iteritems():
                key = key.replace('_', '-')
                if isinstance(value, bool):
                    cache_control.append(key)
                elif isinstance(value, basestring):
                    cache_control.append('%s="%s"' % (key, value))
                else:
                    cache_control.append('%s=%s' % (key, value))
            cache_control = ', '.join(cache_control)

            response.headers['Cache-Control'] = cache_control
            return response
        return wrapper
    return decorator



#!/usr/bin/env python
from setuptools import setup

# Adding this import avoids an exception caused by nosetests
# http://stackoverflow.com/questions/9352656/python-assertionerror-when-running-nose-tests-with-coverage
from multiprocessing import util
util = util  # make pyflakes/etc happy


setup(
    name='clay-flask',
    version='2.1.1',
    author='Jeremy Grosser',
    author_email='jeremy@uber.com',
    packages=['clay'],
    description='Clay is a framework for building RESTful backend services using best practices.',
    install_requires=[
        'flask',
    ],
    tests_require=[
        'nose',
        'webtest',
        'mock >= 1.0.0',
    ],
    test_suite='nose.collector',
    entry_points={
        'console_scripts': [
            'clay-devserver = clay.server:devserver',
            'clay-celery = clay.celery:main',
        ],
    },
)



from __future__ import absolute_import
from clay.server import application
from clay import config

log = config.get_logger('clay.wsgi')

views = config.get('views', [])
if not views:
    log.warning('No clay view modules configured')

for modulename in views:
    log.debug('Loading views from %s' % modulename)
    module = __import__(modulename)



from __future__ import absolute_import
from celery import Celery
from clay import config


log = config.get_logger('clay.celery')

celery = Celery(log=log)
celery.config_from_object(config.get('celery'))


def main():
    '''
    Run a celery worker process
    '''
    celery.worker_main()


if __name__ == '__main__':
    main()



from __future__ import absolute_import

import functools
import socket
import time

from clay import config

log = config.get_logger('clay.stats')


class StatsConnection(object):
    '''
    Handles the lifecycle of stats sockets and connections.
    '''
    def __init__(self):
        self.sock = None
        self.proto = None
        self.host = None
        self.port = None

    def __str__(self):
        if self.sock is not None:
            return 'StatsConnection %s %s:%i (connected)' % (
                   self.proto, self.host, self.port)
        else:
            return 'StatsConnection %s %s:%i (not connected)' % (
                   self.proto, self.host, self.port)

    def get_socket(self):
        '''
        Creates and connects a new socket, or returns an existing one if this
        method was called previously. Returns a (protocol, socket) tuple, where
        protocol is either 'tcp' or 'udp'. If the returned socket is None, the
        operation failed and details were logged.
        '''
        if self.sock is not None:
            return (self.proto, self.sock)

        proto = config.get('statsd.protocol', 'udp')
        self.proto = proto
        self.host = config.get('statsd.host', None)
        self.port = config.get('statsd.port', 8125)

        if proto == 'udp':
            self.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
            log.debug('Created udp statsd socket')
            return (proto, self.sock)

        if proto == 'tcp':
            if self.host is None or not isinstance(self.port, int):
                log.error('Invalid TCP statsd config: host=%r port=%r',
                          self.host, self.port)
                self.sock = None
            else:
                try:
                    self.sock = socket.create_connection(address=(self.host, self.port), timeout=4.0)
                    log.debug('Connected tcp statsd socket to %s:%i',
                              self.host, self.port)
                except socket.error:
                    log.exception('Cannot open tcp stats socket %s:%i',
                                  self.host, self.port)
                    self.sock = None
            return (proto, self.sock)

        log.warning('Unknown protocol configured for statsd socket: %s', proto)
        return (proto, None)

    def reset(self):
        '''
        Close and remove references to the socket.
        '''
        if self.sock is None:
            return
        try:
            self.sock.close()
        except socket.error:
            pass
        self.sock = None
        log.debug('Reset statsd socket')

    def send(self, stat):
        '''
        Send a raw stat line to statsd. A new socket will be opened and
        connected if necessary. Returns True if the stat was sent successfully.

        :param stat: The stat to be sent to statsd, with no trailing newline
        :type stat: string
        :rtype: boolean
        '''
        proto, sock = self.get_socket()
        if sock is None:
            return False
        
        if not stat.endswith('\n'):
            stat += '\n'

        try:
            if proto == 'udp':
                sock.sendto(stat, 0, (self.host, self.port))
                return True

            if proto == 'tcp':
                sock.sendall(stat)
                return True
        except socket.error:
            log.exception('Unable to send to statsd, resetting socket')
            self.reset()
        return False

connection = StatsConnection()
send = connection.send  # backwards compatibility


class Timer(object):
    '''
    Context manager for recording wall-clock timing stats.

    with clay.stats.Timer("myapp.example"):
        # do some work
    '''
    def __init__(self, key):
        self.key = key
        self.start = None

    def __enter__(self):
        self.start = time.clock()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        now = time.clock()
        elapsed_ms = ((now - self.start) * 1000.0)
        timing(self.key, elapsed_ms)


def count(key, n, sample=1.0):
    '''
    Increment a counter by n, or decrement if n is negative.

    :param key: The key to increment by the count
    :type key: string
    :param n: The number to increment by
    :type n: integer
    :param sample: Optional sample rate to scale the counter by. Must be a
                   float between 0.0 and 1.0. Defaults to 1.0
    :type sample: float
    '''
    if sample == 1.0:
        return connection.send('%s:%i|c' % (key, n))
    else:
        return connection.send('%s:%i|c|@%f' % (key, n, sample))


def timing(key, ms):
    '''
    Send a timing stat to statsd

    :param key: A key identifying this stat
    :type key: string
    :param ms: A floating point number of milliseconds
    :type ms: float
    '''
    if not isinstance(ms, float):
        ms = float(ms)
    return connection.send('%s:%f|ms' % (key, ms))


def gauge(key, value):
    '''
    Send an instantaneous gauge value to statsd

    :param key: Name of this gauge
    :type key: string
    :param value: Gauge value or delta
    :type value: float
    '''
    if not isinstance(value, float):
        value = float(value)
    return connection.send('%s:%f|g' % (key, value))


def unique_set(key, value):
    '''
    Send a set stat to statsd, counting the approximate number of unique
    key/value pairs.

    :param key: Name of this set
    :type key: string
    :param value: Set value
    :type value: string
    '''
    return connection.send('%s:%s|s' % (key, value))


def wrapper(prefix):
    '''
    Decorator that logs timing, call count, and exception count statistics to
    statsd. Given a prefix of "example", the following keys would be created:

    stats.counts.example.calls
    stats.counts.example.exceptions
    stats.timers.example.duration

    :param: prefix
    :type key: Prefix for stats keys to be created under
    '''

    def clay_stats_wrapper(func):
        @functools.wraps(func)
        def wrap(*args, **kwargs):
            count('%s.calls' % prefix, 1)
            try:
                with Timer('%s.duration' % prefix):
                    return func(*args, **kwargs)
            except Exception:
                count('%s.exceptions' % prefix, 1)
                raise
        return wrap
    return clay_stats_wrapper



from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import smtplib

from clay import config

log = config.get_logger('clay.mail')


def _string_or_list(obj):
    '''
    If obj is a string, it's converted to a single element list, otherwise
    it's just returned as-is under the assumption that it's already a list. No
    further type checking is performed.
    '''

    if isinstance(obj, basestring):
        return [obj]
    else:
        return obj


def sendmail(mailto, subject, message, subtype='html', charset='utf-8', **headers):
    '''
    Send an email to the given address. Additional SMTP headers may be specified
    as keyword arguments.
    '''

    # mailto arg is explicit to ensure that it's always set, but it's processed
    # mostly the same way as all other headers
    headers['To'] = _string_or_list(mailto)

    msg = MIMEMultipart('alternative')
    msg['Subject'] = subject
    for key, value in headers.iteritems():
        for val in _string_or_list(value):
            msg.add_header(key, val)

    text = MIMEText(message, subtype, charset)
    msg.attach(text)

    if not 'From' in msg:
        # we support both smtp.from and mail.from for legacy reasons
        # smtp.from is the correct usage.
        msg['From'] = config.get('smtp.from') or config.get('mail.from')
    mailfrom = msg['From']
    assert isinstance(mailfrom, basestring)

    recipients = []
    for toheader in ('To', 'CC', 'BCC'):
        recipients += msg.get_all(toheader, [])
    if 'BCC' in msg:
        del msg['BCC']

    smtp = smtplib.SMTP(config.get('smtp.host'), config.get('smtp.port'))
    if config.get('smtp.username') and config.get('smtp.password'):
        smtp.login(config.get('smtp.username'), config.get('smtp.password'))
    smtp.sendmail(mailfrom, recipients, msg.as_string())
    smtp.quit()
    log.info('Sent email to %s (Subject: %s)', recipients, subject)
